"""Asynchronous gRPC client support."""
import asyncio
import logging
from types import ModuleType
from typing import Any, ClassVar, Dict, Optional, Sequence, Type

import grpc.aio
from google.protobuf.empty_pb2 import Empty
from google.protobuf.message import Message
from grpc import ChannelCredentials
from grpc.aio import Channel

from .constants import GRPC_MAX_RECONNECT_BACKOFF_MS, GRPC_MIN_RECONNECT_BACKOFF_MS
from .decoder import GrpcDecoder
from .exceptions import ServiceNotFound
from .utils import format_message, get_call_id

_LOGGER = logging.getLogger(__name__)


class AsyncGrpcClient:
    """Base class for asynchronous gRPC clients.

    Class Attributes:
        decoder_cls: A default decoder class. No arguments are provided for its instantiation.
        grpc_modules: A default list of generated _pb2_grpc modules with service stubs.

    Attributes:
        decoder: The decoder.
        grpc_modules: The list of generated _pb2_grpc modules with service stubs.
    """

    decoder_cls: ClassVar[Type[GrpcDecoder]] = GrpcDecoder
    grpc_modules: Sequence[ModuleType] = ()

    def __init__(
        self, netloc: str, decoder: GrpcDecoder = None, grpc_modules: Sequence[ModuleType] = None,
        credentials: ChannelCredentials = None
    ) -> None:
        """Initialize GRPC backend client instance.

        Args:
            netloc: Network location of a gRPC server.
            decoder: Custom decoder instance.
            grpc_modules: Custom list of generated _pb2_grpc modules with service stubs.
            credentials: Credentials for a secure channel connection. If None, insecure channel is used.
        """
        self.netloc = netloc
        self.decoder = decoder or self.decoder_cls()
        self.grpc_modules = grpc_modules or self.grpc_modules
        self.credentials = credentials
        self._channel: Optional[Channel] = None

        # Stubs cache
        self._stubs: Dict[str, Any] = {}

    @property
    def channel(self) -> Channel:
        """Create channel to GPRC server."""
        if self._channel is not None:
            return self._channel

        options = (
            ('grpc.min_reconnect_backoff_ms', GRPC_MIN_RECONNECT_BACKOFF_MS),
            ('grpc.max_reconnect_backoff_ms', GRPC_MAX_RECONNECT_BACKOFF_MS),
        )
        if self.credentials is not None:
            self._channel = grpc.aio.secure_channel(self.netloc, self.credentials, options=options)
        else:
            self._channel = grpc.aio.insecure_channel(self.netloc, options=options)
        return self._channel

    def get_stub(self, service: str) -> Any:
        """Return stub for the service.

        Stubs are cached, so for each service, there is only one stub created.
        However, stubs are created from channel and have to be reinstantiated
        after reconnect.
        """
        stub_name = '{}Stub'.format(service)

        if stub_name in self._stubs:
            return self._stubs[stub_name]

        for module in self.grpc_modules:
            if hasattr(module, stub_name):
                stub_cls = getattr(module, stub_name)
                self._stubs[stub_name] = stub_cls(self.channel)
                return self._stubs[stub_name]
        else:
            raise ServiceNotFound('Service {} could not be found'.format(stub_name))

    async def call(
        self, service: str, method: str, request: Message,
        *, timeout: float = None, max_retries: int = 10
    ) -> Any:
        """Call grpc method.

        Args:
            service: Name of service, that is used to get service stub.
            method: Name of service method.
            request: Request message passed as an argument to service method.
            blocking: Determines whether to use blocking or non-blocking call.
            timeout: An optional duration in seconds allowed for the gRPC call.
            max_retries: A maximum number of calls to be made.
        """
        call_id = get_call_id()
        _LOGGER.debug('[%s] %s(%s)', call_id, method, format_message(request))

        retries = 0
        try:
            while True:
                await asyncio.sleep(0.1 * retries)  # Wait for increasing amount of time before each call
                try:
                    retries += 1
                    stub = self.get_stub(service)
                    grpc_method = getattr(stub, method)
                    reply = await grpc_method(request, timeout=timeout)
                    break
                except grpc.RpcError as error:
                    if error.code() == grpc.StatusCode.UNAVAILABLE:
                        if retries < max_retries:
                            _LOGGER.info('[%s] Connection failed. Attempting retry.', call_id)
                        else:
                            _LOGGER.info('[%s] Connection failed. Retries were unsuccessful.', call_id)
                            raise
                    else:
                        raise

        except grpc.RpcError as error:
            _LOGGER.info('[%s] %s failed with %s', call_id, method, error)
            raise

        _LOGGER.debug('[%s] %s returned %s', call_id, method, format_message(reply))
        return self.decode_reply(reply)

    def decode_reply(self, reply: Message) -> Optional[Dict[str, Any]]:
        """Decode reply and return decoded result or raise decoded exception.

        Reply must either be Empty or contain at least one of fields `data` and `exception`.
        """
        # Assert the supported structure of messages
        assert ('exception' in reply.DESCRIPTOR.fields_by_name or 'data' in reply.DESCRIPTOR.fields_by_name  # nosec
                or isinstance(reply, Empty))

        # Decode exception first
        if 'exception' in reply.DESCRIPTOR.fields_by_name and reply.HasField('exception'):
            raise self.decoder.decode_exception(reply.exception)  # type: ignore

        # Decode data otherwise
        if 'data' in reply.DESCRIPTOR.fields_by_name and reply.HasField('data'):
            return self.decoder.decode(reply.data)  # type: ignore

        # The reply is empty
        return None
